from typing import Literal

from .base import MercadonaItem, lazy_load_property
from .category import Category
from ..constants import *
from ..utils.urls import get_file_path
from .photo import Photo


def require_complete_data(func):
    def wrapper(self):
        if self._is_data_incomplete():
            self._fetch_data()

        value = func(self)
        return value

    return wrapper


class Product(MercadonaItem):
    def __init__(
        self,
        id: str | dict,
        warehouse: str = "mad1",
        language: Literal["es", "en"] = "es",
    ):
        if isinstance(id, dict):
            endpoint = f"/api/products/{id.get("id")}/"
        else:
            endpoint = f"/api/products/{id}/"

        super().__init__(id, endpoint, warehouse, language)

    def _is_data_incomplete(self):
        if self._data is None:
            return True
        
        details = self._data.get("details", {})
        return not details

    @lazy_load_property
    @require_complete_data
    def ean(self) -> str:
        """
        Returns the European Article Number (EAN) for the product.
        """
        return self._data.get("ean")

    @lazy_load_property
    def name(self) -> str:
        """
        Returns the display name of the product.
        """
        return self._data.get("display_name")

    @lazy_load_property
    def slug(self) -> str:
        """
        Returns the slug (URL-friendly version of the name) for the product.
        """
        return self._data.get("slug")

    @lazy_load_property
    @require_complete_data
    def legal_name(self) -> str | None:
        """
        Returns the legal name of the product.
        """
        details = self._data.get("details", {})
        return details.get("legal_name", None)

    @lazy_load_property
    def unit_price(self) -> float | None:
        """
        Returns the unit price of the product.
        """
        return float(self._data.get("price_instructions", {}).get("unit_price"))

    @lazy_load_property
    def bulk_price(self) -> float | None:
        """
        Returns the bulk price of the product.

        Returns:
            float: The bulk price as a float, or None if not available.
        """
        return float(self._data.get("price_instructions", {}).get("bulk_price", None))

    @lazy_load_property
    def is_discounted(self) -> bool:
        """
        Checks if the product is currently discounted.
        """
        return self._data.get("price_instructions", {}).get("price_decreased")

    @lazy_load_property
    @require_complete_data
    def previous_price(self) -> float | None:
        """
        Returns the previous unit price before any discounts.
        """
        return self._data.get("price_instructions", {}).get("previous_unit_price")

    @lazy_load_property
    def iva(self) -> int:
        """
        Returns the tax rate (IVA) applied to the product.
        """
        return self._data.get("price_instructions", {}).get("iva")

    @lazy_load_property
    def age_check(self) -> bool:
        """
        Checks if the product requires an age check.
        """
        return self._data.get("badges", {}).get("requires_age_check", False)

    @lazy_load_property
    @require_complete_data
    def alcohol_by_volume(self) -> float | None:
        """
        Returns the alcohol by volume percentage of the product.
        """
        details = self._data.get("details", {})
        percentage = details.get("alcohol_by_volume")

        if percentage:
            return float(percentage.removesuffix("º"))

    @lazy_load_property
    def is_new(self) -> bool:
        """
        Checks if the product is new.
        """
        return self._data.get("price_instructions", {}).get("is_new", False)

    @lazy_load_property
    def is_pack(self) -> bool:
        """
        Checks if the product is sold as a pack.
        """
        return self._data.get("price_instructions", {}).get("is_pack", False)

    @lazy_load_property
    def pack_size(self) -> int | None:
        """
        Returns the pack size of the product, if applicable.
        """
        if not self.is_pack:
            return None

        return self._data.get("price_instructions", {}).get("pack_size",  None)

    @lazy_load_property
    def total_units(self) -> int | None:
        """
        Returns the total units in the pack, if applicable.
        """
        if not self.is_pack:
            return None
        
        return self._data.get("price_instructions", {}).get("total_units", None)

    @lazy_load_property
    @require_complete_data
    def photos(self) -> list[Photo]:
        """
        Returns the list of photos for the product.
        """
        photos = self._data.get("photos", [])
        return [Photo(get_file_path(p.get("regular"))) for p in photos]

    @lazy_load_property
    @require_complete_data
    def description(self) -> str:
        """
        Returns the description of the product.
        """
        details = self._data.get("details", {})
        return details.get("description", "")

    @lazy_load_property
    def minimum_amount(self) -> int:
        """
        Returns the minimum amount that can be purchased.
        """
        return int(self._data.get("price_instructions", {}).get("min_bunch_amount", 1))

    @lazy_load_property
    def weight(self) -> float:
        """
        Returns the weight of the product.
        """
        return self._data.get("price_instructions", {}).get("unit_size")

    @lazy_load_property
    def brand(self) -> str:
        """
        Returns the brand of the product.
        """
        return self._data.get("brand")

    @lazy_load_property
    @require_complete_data
    def origin(self) -> str | None:
        """
        Returns the origin of the product.
        """
        details = self._data.get("details", {})
        return details.get("origin", None)

    @lazy_load_property
    @require_complete_data
    def supplier(self) -> str | None:
        """
        Returns the supplier of the product.
        """
        details = self._data.get("details", {})
        suppliers = details.get("suppliers", [])

        if suppliers:
            return suppliers[0].get("name")

    @lazy_load_property
    def category(self) -> list[str]:
        """
        Returns the category of the product.
        """
        high_level_category = self._data.get("categories", [])[0]
        category_data = high_level_category.get("categories", [])[0]
        
        category = Category(category_data, self.warehouse, self.language)
        return category
    
    def __dict__(self):
        """
        Converts the product object to a dictionary.
        """
        return {
            "id": self.id,
            "warehouse": self.warehouse,
            "language": self.language,
            "ean": self.ean,
            "name": self.name,
            "slug": self.slug,
            "legal_name": self.legal_name,
            "unit_price": self.unit_price,
            "bulk_price": self.bulk_price,
            "is_discounted": self.is_discounted,
            "previous_price": self.previous_price,
            "iva": self.iva,
            "age_check": self.age_check,
            "alcohol_by_volume": self.alcohol_by_volume,
            "is_new": self.is_new,
            "is_pack": self.is_pack,
            "pack_size": self.pack_size,
            "description": self.description,
            "minimum_amount": self.minimum_amount,
            "weight": self.weight,
            "brand": self.brand,
            "origin": self.origin,
            "supplier": self.supplier,
            "category": self.category.name
        }



from dataclasses import dataclass
from typing import Literal
import os, requests

from ..utils.urls import *


@dataclass
class Photo:
    """
    Represents a product photo.

    Args:
        file_name (str): The file name corresponding to the image (e.g. "cea11c6ef934dff6c6a018df3b757b8d.jpg")
    """

    file_name: str

    def __post_init__(self):
        if is_url(self.file_name):
            self.file_name = get_file_path(self.file_name)

        self.url = f"https://prod-mercadona.imgix.net/images/{self.file_name}"

    def get_size(
        self, width: int, height: int, fit_mode: Literal["crop", "fit"] = "crop"
    ) -> str:
        """
        Generates a resized image URL based on specified width, height, and fit mode.

        Args:
            width (int): Desired width of the image.
            height (int): Desired height of the image.
            fit_mode (str, optional): Fit mode for resizing, default is 'crop'.

        Returns:
            str: Resized image URL.
        """
        return f"{self.url}?fit={fit_mode}&h={height}&w={width}"

    def save(self, path: str, width: int = None, height: int = None, fit_mode="crop"):
        """
        Downloads and saves the photo to a specified path.

        Args:
            path (str): Path where the photo will be saved.
            width (int, optional): Desired width of the downloaded photo.
            height (int, optional): Desired height of the downloaded photo.
            fit_mode (str, optional): Fit mode for resizing, default is 'crop'.
        """
        photo_url = (
            self.get_size(width, height, fit_mode) if width and height else self.url
        )
        try:
            response = requests.get(photo_url)
            response.raise_for_status()

            # Ensure the directory exists
            directory = os.path.dirname(path)
            if directory:
                os.makedirs(directory, exist_ok=True)

            with open(path, "wb") as file:
                file.write(response.content)
        except requests.exceptions.RequestException as e:
            print(f"Failed to download the photo: {e}")



from dataclasses import dataclass, field
from typing import Literal
from urllib.parse import urljoin
import time

from ..utils.api import fetch_json
from ..constants import API_URL


def lazy_load_property(func):
    @property
    def wrapper(self):
        if self._is_empty():
            self._fetch_data()

        if self.not_found():
            return None

        return func(self)

    return wrapper


@dataclass
class MercadonaItem:
    """
    Represents a season in Mercadona's catalog.

    Args:
        id (str | dict): Season identifier.
        endpoint (str): API endpoint to fetch data from.
        warehouse (str): Warehouse or distribution center postal code.
        language (str): Language for the API response. Defaults to "es".
    """

    id: str | dict
    endpoint: str = field(repr=False)
    warehouse: str = "mad1"
    language: Literal["es", "en"] = field(default="es", init=True, repr=False)

    def __post_init__(self):
        self._data = {}

        if isinstance(self.id, dict):
            self._data = self.id
            self.id = str(self._data.get("id"))

    def not_found(self):
        if self._is_empty():
            self._fetch_data()

        return bool(self._data.get("err_code"))

    def _fetch_with_context(self, endpoint: str) -> dict:
        url = urljoin(API_URL, endpoint)
        return fetch_json(url, {"lang": self.language, "wh": self.warehouse})

    def _fetch_data(self, retry_attempts: int = 3, retry_delay: int = 20) -> None:
        attempt = 0
        while attempt <= retry_attempts:
            self._data = self._fetch_with_context(self.endpoint)
            err_code = self._data.get("err_code")

            if err_code is None:
                # No error, data fetched successfully
                break
            elif err_code == 429:
                # Too Many Requests, retry
                attempt += 1
                if attempt > retry_attempts:
                    print(
                        f"Exceeded maximum retry attempts ({retry_attempts}) for {self}."
                    )
                    break
                calculated_delay = attempt**2 * retry_delay
                time.sleep(calculated_delay)
            elif err_code == 404:
                # Not Found, no need to retry
                print(f"Couldn't find {self} :(")
                break
            else:
                # Other errors, stop retrying
                print(f"Error fetching data for {self}.")
                break

    def _is_empty(self):
        return not bool(self._data)

    def __dict__(self):
        if self._is_empty():
            self._fetch_data()

        return self._data



from urllib.parse import urlparse
import os


def is_url(string):
    parsed_url = urlparse(string)
    return parsed_url.scheme in ("http", "https", "ftp")


def get_file_path(url):
    return os.path.basename(urlparse(url).path)



import requests


def get_warehouse_code(postal_code):
    """
    Get warehouse code for a given postal code.

    Args:
        postal_code (str): The postal code to query.

    Returns:
        str or None: Warehouse code if found, None otherwise.
    """
    url = "https://tienda.mercadona.es/api/postal-codes/actions/change-pc/"
    payload = {"new_postal_code": postal_code}
    headers = {"Content-Type": "application/json"}

    try:
        response = requests.put(url, json=payload, headers=headers)
        if response.status_code == 200:
            return response.headers.get("X-Customer-Wh")
    except Exception as e:
        print(f"Error for postal code {postal_code}: {e}")
    return None



from typing import Literal

from .base import MercadonaItem, lazy_load_property

def require_complete_data(func):
    def wrapper(self):
        if self._is_data_incomplete():
            self._fetch_data()

        value = func(self)
        return value

    return wrapper

class Category(MercadonaItem):
    def __init__(
        self,
        id: str | dict,
        warehouse: str,
        language: Literal["es", "en"] = "es",
    ):
        if isinstance(id, dict):
            endpoint = f"/api/categories/{id.get("id")}/"
        else:
            endpoint = f"/api/categories/{id}/"

        super().__init__(id, endpoint, warehouse, language)

    def _is_data_incomplete(self):      
        subcategories = self._data.get("categories", None)
        return subcategories is None

    @lazy_load_property
    @require_complete_data
    def products(self):
        from .product import Product

        category_products = []
        subcategories = self._data.get("categories", [])

        for subcategory in subcategories:
            products = subcategory.get("products", None)

            for product_data in products:
                product = Product(product_data, self.warehouse, self.language)
                category_products.append(product)

        return category_products

    @lazy_load_property
    def name(self):
        return self._data.get("name")



from setuptools import setup, find_packages
from codecs import open
from os import path

# Read version from VERSION file
here = path.abspath(path.dirname(__file__))

VERSION = "1.0.3"
DESCRIPTION = (
    "A Mercadona interface for Python to track product prices, amounts, and more."
)

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="mercapy",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Joel Taylor",
    author_email="contact@joeltaylor.business",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    keywords=["mercadona", "api", "sdk", "data science", "prices", "information"],
    packages=find_packages(exclude=["docs", "tests"]),
    install_requires=["requests"],
    setup_requires=["setuptools>=38.6.0"],
)



from typing import Literal

from .base import MercadonaItem, lazy_load_property
from .product import Product


class Season(MercadonaItem):
    def __init__(
        self,
        id: str | dict,
        warehouse: str,
        language: Literal["es", "en"] = "es",
    ):
        if isinstance(id, dict):
            endpoint = f"/api/home/seasons/{id.get("id")}/"
        else:
            endpoint = f"/api/home/seasons/{id}/"
        
        super().__init__(id, endpoint, warehouse, language)

    @lazy_load_property
    def title(self) -> str:
        return self._data.get("title")

    @lazy_load_property
    def products(self) -> list[Product]:
        items = self._data.get("items", [])
        return [Product(i, self.warehouse, self.language) for i in items]



from typing import Literal
from urllib.parse import urljoin

from .constants import WAREHOUSES
from .utils.warehouses import get_warehouse_code
from .utils.api import *
from .elements import Product, Season, Category


class Mercadona:

    def __init__(
        self,
        postcode: str,
        language: Literal["es", "en"] = "es",
    ) -> None:
        """
        Represents a Mercadona warehouse, from where their catalog can browsed.
        A postcode that isn't a Mercadona warehouse will be treated as a city postcode, and will find the closes warehouse to the population if any.

        Args:
            postcode (str): The postcode from where products are being accessed. From there, the closest warehouse will be found. Warehouse codes are accepted too (e.g. "mad1", "vlc1", etc.)
            language (str): The language of the information recieved. Defaults to "es": Spanish. Can also be "en": English.
        """
        self.language = language

        if postcode in WAREHOUSES:
            self.postcode = postcode
            self.warehouse = self.postcode
        else:
            self.postcode = postcode
            self.warehouse = get_warehouse_code(self.postcode)

    def _get_with_context(self, url: str):
        return fetch_json(url, {"lang": self.language, "wh": self.warehouse})

    def search(self, query: str) -> list[Product]:
        """
        Queries Mercadona's products using their provider "Algolia".

        Args:
            query (str): Search query (e.g. Dish Soap).
            lang (str): Language code. Defaults to spanish. Can be "en" too.

        Rerturns:
            list[Product]: List of products related to the search.
        """
        response = query_algolia(query, self.warehouse, self.language)
        hits = response.get("hits", [])

        products = []
        for h in hits:
            product = Product(h["id"], self.warehouse, self.language)
            products.append(product)

        return products

    def get_home_recommendations(self) -> dict:
        """
        Retrieves product recommendations for the home page grouped by sections.

        Returns:
            dict: Dictionary where keys are layout names and values are lists of recommended products. The lists of products can also include banners which often are Season objects.
        """
        url = urljoin(API_URL, f"/api/home/")
        response = self._get_with_context(url)

        sections = response.get("sections", [])

        # Dictionary to store products grouped by sections
        section_products = {}

        for section in sections:
            section_name = section.get("layout")

            items = section.get("content", {}).get("items", [])

            for item in items:
                if item.get("bg_colors", None):
                    parsed_item = Season(str(item["id"]), self.warehouse, self.language)
                else:
                    parsed_item = Product(item, self.warehouse, self.language)

                if section_products.get(section_name, None):
                    section_products[section_name].append(parsed_item)
                else:
                    section_products[section_name] = [parsed_item]

        return section_products

    def get_new_arrivals(self) -> list[Product]:
        """
        New product arrivals at Mercadona

        Args:
            lang (str): Language code. Defaults to spanish. Can be "en" too.

        Returns:
            list[Product]: List of new product arrivals.
        """
        url = urljoin(API_URL, f"/api/home/new-arrivals/")
        response = self._get_with_context(url)

        products = []
        for item in response.get("items", []):
            product = Product(item, self.warehouse, self.language)
            products.append(product)

        return products

    def get_categories(self) -> list[Category]:
        url = urljoin(API_URL, "/api/categories/")
        response = self._get_with_context(url)

        # Get all level 1 categories
        lvl1_categories = []

        results = response.get("results", [])
        for result in results:
            categories = result.get("categories", [])
            for c in categories:
                category = Category(c, self.warehouse, self.language)
                lvl1_categories.append(category)

        return lvl1_categories

    def get_catalog(self) -> list[Product]:
        return [p for c in self.get_categories() for p in c.products]



import requests, json
from ..constants import *


def fetch_json(url: str, params: dict = None) -> dict:
    """
    Fetches JSON data from a given URL.

    Args:
        url (str): The URL to fetch data from.
        params (dict, optional): The parameters to send with the request. Defaults to None.

    Returns:
        dict: The JSON response as a dictionary, or an empty dictionary if there's an error.
    """
    try:
        response = requests.get(url, params=params, allow_redirects=False)
        response.raise_for_status()

        return response.json()
    except requests.exceptions.RequestException as e:
        return {"err_code": response.status_code, "err_message": e}


def query_algolia(query: str, warehouse: str, lang: str = "es") -> dict | None:
    """
    Queries Algolia for product data.

    Args:
        query (str): The query string.
        lang (str, optional): The language for the query. Defaults to "es".

    Returns:
        dict or None: The JSON response as a dictionary, or None if there's an error.
    """
    url = f"https://7uzjkl1dj0-dsn.algolia.net/1/indexes/products_prod_{warehouse}_{lang}/query"

    # Headers required for the request
    headers = {
        "x-algolia-application-id": ALGOLIA_APP_ID,
        "x-algolia-api-key": ALGOLIA_API_KEY,
        "Content-Type": "application/json",
    }

    # Data payload for the request
    payload = {"params": f"query={query}"}

    try:
        with requests.post(url, headers=headers, data=json.dumps(payload)) as response:
            response.raise_for_status()

            # Check if the request was successful
            if response.ok:
                return response.json()

    except requests.exceptions.RequestException as e:
        print(f"An error occurred: {e}")

    return None
